@@ -11,7 +11,7 @@
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
-# obtain one at http://mozilla.org/MPL/2.0/.
+# obtain one at https://mozilla.org/MPL/2.0/.
 #
 # END HEADER
 
@@ -24,10 +24,10 @@ from weakref import WeakKeyDictionary
 import attr
 
 from hypothesis import HealthCheck, Phase, Verbosity, settings as Settings
-from hypothesis._settings import local_settings, note_deprecation
+from hypothesis._settings import local_settings
+from hypothesis.internal.cache import LRUReusedCache
 from hypothesis.internal.compat import (
     Counter,
-    benchmark_time,
     ceil,
     hbytes,
     hrange,
@@ -38,22 +38,22 @@ from hypothesis.internal.compat import (
 from hypothesis.internal.conjecture.data import (
     MAX_DEPTH,
     ConjectureData,
+    Overrun,
     Status,
     StopTest,
 )
+from hypothesis.internal.conjecture.datatree import DataTree
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.reporting import debug_report
 
 # Tell pytest to omit the body of this module from tracebacks
-# http://doc.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
+# https://docs.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
 __tracebackhide__ = True
 
 
-HUNG_TEST_TIME_LIMIT = 5 * 60
 MAX_SHRINKS = 500
-
-CACHE_RESET_FREQUENCY = 1000
+CACHE_SIZE = 10000
 MUTATION_POOL_SIZE = 100
 
 
@@ -68,7 +68,6 @@ class HealthCheckState(object):
 class ExitReason(Enum):
     max_examples = 0
     max_iterations = 1
-    timeout = 2
     max_shrinks = 3
     finished = 4
     flaky = 5
@@ -86,7 +85,6 @@ class ConjectureRunner(object):
         self.call_count = 0
         self.event_call_counts = Counter()
         self.valid_examples = 0
-        self.start_time = benchmark_time()
         self.random = random or Random(getrandbits(128))
         self.database_key = database_key
         self.status_runtimes = {}
@@ -106,67 +104,18 @@ class ConjectureRunner(object):
         self.health_check_state = None
 
         self.used_examples_from_database = False
-        self.reset_tree_to_empty()
+        self.tree = DataTree(cap=self.cap)
 
-    def reset_tree_to_empty(self):
-        # Previously-tested byte streams are recorded in a prefix tree, so that
-        # we can:
-        # - Avoid testing the same stream twice (in some cases).
-        # - Avoid testing a prefix of a past stream (in some cases),
-        #   since that should only result in overrun.
-        # - Generate stream prefixes that we haven't tried before.
-
-        # Tree nodes are stored in an array to prevent heavy nesting of data
-        # structures. Branches are dicts mapping bytes to child nodes (which
-        # will in general only be partially populated). Leaves are
-        # ConjectureData objects that have been previously seen as the result
-        # of following that path.
-        self.tree = [{}]
-
-        # A node is dead if there is nothing left to explore past that point.
-        # Recursively, a node is dead if either it is a leaf or every byte
-        # leads to a dead node when starting from here.
-        self.dead = set()
-
-        # We rewrite the byte stream at various points during parsing, to one
-        # that will produce an equivalent result but is in some sense more
-        # canonical. We keep track of these so that when walking the tree we
-        # can identify nodes where the exact byte value doesn't matter and
-        # treat all bytes there as equivalent. This significantly reduces the
-        # size of the search space and removes a lot of redundant examples.
-
-        # Maps tree indices where to the unique byte that is valid at that
-        # point. Corresponds to data.write() calls.
-        self.forced = {}
-
-        # Maps tree indices to a mask that restricts bytes at that point.
-        # Currently this is only updated by draw_bits, but it potentially
-        # could get used elsewhere.
-        self.masks = {}
-
-        # Where a tree node consists of the beginning of a block we track the
-        # size of said block. This allows us to tell when an example is too
-        # short even if it goes off the unexplored region of the tree - if it
-        # is at the beginning of a block of size 4 but only has 3 bytes left,
-        # it's going to overrun the end of the buffer regardless of the
-        # buffer contents.
-        self.block_sizes = {}
+        # We want to be able to get the ConjectureData object that results
+        # from running a buffer without recalculating, especially during
+        # shrinking where we need to know about the structure of the
+        # executed test case.
+        self.__data_cache = LRUReusedCache(CACHE_SIZE)
 
     def __tree_is_exhausted(self):
-        return 0 in self.dead
+        return self.tree.is_exhausted
 
     def test_function(self, data):
-        if benchmark_time() - self.start_time >= HUNG_TEST_TIME_LIMIT:
-            fail_health_check(
-                self.settings,
-                (
-                    "Your test has been running for at least five minutes. This "
-                    "is probably not what you intended, so by default Hypothesis "
-                    "turns it into an error."
-                ),
-                HealthCheck.hung_test,
-            )
-
         self.call_count += 1
         try:
             self._test_function(data)
@@ -206,111 +155,12 @@ class ConjectureRunner(object):
         #    and these will tend to have a very high duplication rate. This is
         #    where the tree usage really shines.
         #
-        # Unfortunately, as well as being the less useful type of tree usage,
-        # the first type is also the most expensive! Once we've entered shrink
-        # mode our time remaining is essentially bounded - we're just here
-        # until we've found the minimal example. In exploration mode, we might
-        # be early on in a very long-running processs, and keeping everything
-        # we've ever seen lying around ends up bloating our memory usage
-        # substantially by causing us to use O(max_examples) memory.
-        #
-        # As a compromise, what we do is reset the cache every so often. This
-        # keeps our memory usage bounded. It has a few unfortunate failure
-        # modes in that it means that we can't always detect when we should
-        # have stopped - if we are exploring a language which has only slightly
-        # more than cache reset frequency number of members, we will end up
-        # exploring indefinitely when we could have stopped. However, this is
-        # a fairly unusual case - thanks to exponential blow-ups in language
-        # size, most languages are either very large (possibly infinite) or
-        # very small. Nevertheless we want CACHE_RESET_FREQUENCY to be quite
-        # high to avoid this case coming up in practice.
-        if (
-            self.call_count % CACHE_RESET_FREQUENCY == 0
-            and not self.interesting_examples
-        ):
-            self.reset_tree_to_empty()
-
-        # First, iterate through the result's buffer, to create the node that
-        # will hold this result. Also note any forced or masked bytes.
-        tree_node = self.tree[0]
-        indices = []
-        node_index = 0
-        for i, b in enumerate(data.buffer):
-            # We build a list of all the node indices visited on our path
-            # through the tree, since we'll need to refer to them later.
-            indices.append(node_index)
-
-            # If this buffer position was forced or masked, then mark its
-            # corresponding node as forced/masked.
-            if i in data.forced_indices:
-                self.forced[node_index] = b
-            try:
-                self.masks[node_index] = data.masked_indices[i]
-            except KeyError:
-                pass
-
-            try:
-                # Use the current byte to find the next node on our path.
-                node_index = tree_node[b]
-            except KeyError:
-                # That node doesn't exist yet, so create it.
-                node_index = len(self.tree)
-                # Create a new branch node. If this should actually be a leaf
-                # node, it will be overwritten when we store the result.
-                self.tree.append({})
-                tree_node[b] = node_index
-
-            tree_node = self.tree[node_index]
-
-            if node_index in self.dead:
-                # This part of the tree has already been marked as dead, so
-                # there's no need to traverse any deeper.
-                break
-
-        # At each node that begins a block, record the size of that block.
-        for u, v in data.all_block_bounds():
-            # This can happen if we hit a dead node when walking the buffer.
-            # In that case we already have this section of the tree mapped.
-            if u >= len(indices):
-                break
-            self.block_sizes[indices[u]] = v - u
-
-        # Forcibly mark all nodes beyond the zero-bound point as dead,
-        # because we don't intend to try any other values there.
-        self.dead.update(indices[self.cap :])
-
-        # Now store this result in the tree (if appropriate), and check if
-        # any nodes need to be marked as dead.
-        if data.status != Status.OVERRUN and node_index not in self.dead:
-            # Mark this node as dead, because it produced a result.
-            # Trying to explore suffixes of it would not be helpful.
-            self.dead.add(node_index)
-            # Store the result in the tree as a leaf. This will overwrite the
-            # branch node that was created during traversal.
-            self.tree[node_index] = data
-
-            # Review the traversed nodes, to see if any should be marked
-            # as dead. We check them in reverse order, because as soon as we
-            # find a live node, all nodes before it must still be live too.
-            for j in reversed(indices):
-                mask = self.masks.get(j, 0xFF)
-                assert _is_simple_mask(mask)
-                max_size = mask + 1
-
-                if len(self.tree[j]) < max_size and j not in self.forced:
-                    # There are still byte values to explore at this node,
-                    # so it isn't dead yet.
-                    break
-                if set(self.tree[j].values()).issubset(self.dead):
-                    # Everything beyond this node is known to be dead,
-                    # and there are no more values to explore here (see above),
-                    # so this node must be dead too.
-                    self.dead.add(j)
-                else:
-                    # Even though all of this node's possible values have been
-                    # tried, there are still some deeper nodes that remain
-                    # alive, so this node isn't dead yet.
-                    break
+        # In aid of this, we keep around just enough of the structure of the
+        # the tree of examples we've seen so far to let us predict whether
+        # something will lead to a known result, and to canonicalize it into
+        # the buffer that would belong to the ConjectureData that you get
+        # from running it.
+        self.tree.add(data)
 
         if data.status == Status.INTERESTING:
             key = data.interesting_origin
@@ -323,35 +173,17 @@ class ConjectureRunner(object):
                 if sort_key(data.buffer) < sort_key(existing.buffer):
                     self.shrinks += 1
                     self.downgrade_buffer(existing.buffer)
+                    self.__data_cache.unpin(existing.buffer)
                     changed = True
 
             if changed:
                 self.save_buffer(data.buffer)
                 self.interesting_examples[key] = data
+                self.__data_cache.pin(data.buffer)
                 self.shrunk_examples.discard(key)
 
             if self.shrinks >= MAX_SHRINKS:
                 self.exit_with(ExitReason.max_shrinks)
-        if (
-            self.settings.timeout > 0
-            and benchmark_time() >= self.start_time + self.settings.timeout
-        ):
-            note_deprecation(
-                (
-                    "Your tests are hitting the settings timeout (%.2fs). "
-                    "This functionality will go away in a future release "
-                    "and you should not rely on it. Instead, try setting "
-                    "max_examples to be some value lower than %d (the number "
-                    "of examples your test successfully ran here). Or, if you "
-                    "would prefer your tests to run to completion, regardless "
-                    "of how long they take, you can set the timeout value to "
-                    "hypothesis.unlimited."
-                )
-                % (self.settings.timeout, self.valid_examples),
-                since="2018-07-29",
-                verbosity=self.settings.verbosity,
-            )
-            self.exit_with(ExitReason.timeout)
 
         if not self.interesting_examples:
             if self.valid_examples >= self.settings.max_examples:
@@ -377,60 +209,7 @@ class ConjectureRunner(object):
         least one novel prefix left to find. If there were not, then the
         test run should have already stopped due to tree exhaustion.
         """
-        prefix = bytearray()
-        node = 0
-        while True:
-            assert len(prefix) < self.cap
-            assert node not in self.dead
-
-            # Figure out the range of byte values we should be trying.
-            # Normally this will be 0-255, unless the current position has a
-            # mask.
-            mask = self.masks.get(node, 0xFF)
-            assert _is_simple_mask(mask)
-            upper_bound = mask + 1
-
-            try:
-                c = self.forced[node]
-                # This position has a forced byte value, so trying a different
-                # value wouldn't be helpful. Just add the forced byte, and
-                # move on to the next position.
-                prefix.append(c)
-                node = self.tree[node][c]
-                continue
-            except KeyError:
-                pass
-
-            # Provisionally choose the next byte value.
-            # This will change later if we find that it was a bad choice.
-            c = self.random.randrange(0, upper_bound)
-
-            try:
-                next_node = self.tree[node][c]
-                if next_node in self.dead:
-                    # Whoops, the byte value we chose for this position has
-                    # already been fully explored. Let's pick a new value, and
-                    # this time choose a value that's definitely still alive.
-                    choices = [
-                        b
-                        for b in hrange(upper_bound)
-                        if self.tree[node].get(b) not in self.dead
-                    ]
-                    assert choices
-                    c = self.random.choice(choices)
-                    node = self.tree[node][c]
-                else:
-                    # The byte value we chose is in the tree, but it still has
-                    # some unexplored descendants, so it's a valid choice.
-                    node = next_node
-                prefix.append(c)
-            except KeyError:
-                # The byte value we chose isn't in the tree at this position,
-                # which means we've successfully found a novel prefix.
-                prefix.append(c)
-                break
-        assert node not in self.dead
-        return hbytes(prefix)
+        return self.tree.generate_novel_prefix(self.random)
 
     @property
     def cap(self):
@@ -538,6 +317,10 @@ class ConjectureRunner(object):
         return b".".join((self.database_key, b"coverage"))
 
     def note_details(self, data):
+        if data.status == Status.OVERRUN:
+            self.__data_cache[data.buffer] = Overrun
+        else:
+            self.__data_cache[data.buffer] = data
         runtime = max(data.finish_time - data.start_time, 0.0)
         self.all_runtimes.append(runtime)
         self.all_drawtimes.extend(data.draw_times)
@@ -792,6 +575,9 @@ class ConjectureRunner(object):
             return
 
         zero_data = self.cached_test_function(hbytes(self.settings.buffer_size))
+        if zero_data.status > Status.OVERRUN:
+            self.__data_cache.pin(zero_data.buffer)
+
         if zero_data.status == Status.OVERRUN or (
             zero_data.status == Status.VALID
             and len(zero_data.buffer) * 2 > self.settings.buffer_size
@@ -811,16 +597,17 @@ class ConjectureRunner(object):
                 HealthCheck.large_base_example,
             )
 
-        # If the language starts with writes of length >= cap then there is
-        # only one string in it: Everything after cap is forced to be zero (or
-        # to be whatever value is written there). That means that once we've
-        # tried the zero value, there's nothing left for us to do, so we
-        # exit early here.
-        for i in hrange(self.cap):
-            if i not in zero_data.forced_indices:
-                break
-        else:
-            self.exit_with(ExitReason.finished)
+        if zero_data is not Overrun:
+            # If the language starts with writes of length >= cap then there is
+            # only one string in it: Everything after cap is forced to be zero (or
+            # to be whatever value is written there). That means that once we've
+            # tried the zero value, there's nothing left for us to do, so we
+            # exit early here.
+            for i in hrange(self.cap):
+                if i not in zero_data.forced_indices:
+                    break
+            else:
+                self.exit_with(ExitReason.finished)
 
         self.health_check_state = HealthCheckState()
 
@@ -917,12 +704,9 @@ class ConjectureRunner(object):
             mutations += 1
 
     def _run(self):
-        self.start_time = benchmark_time()
-
         self.reuse_existing_examples()
         self.generate_new_examples()
         self.shrink_interesting_examples()
-
         self.exit_with(ExitReason.finished)
 
     def shrink_interesting_examples(self):
@@ -1000,61 +784,6 @@ class ConjectureRunner(object):
     def new_shrinker(self, example, predicate):
         return Shrinker(self, example, predicate)
 
-    def prescreen_buffer(self, buffer):
-        """Attempt to rule out buffer as a possible interesting candidate.
-
-        Returns False if we know for sure that running this buffer will not
-        produce an interesting result. Returns True if it might (because it
-        explores territory we have not previously tried).
-
-        This is purely an optimisation to try to reduce the number of tests we
-        run. "return True" would be a valid but inefficient implementation.
-        """
-
-        # Traverse the tree, to see if we have already tried this buffer
-        # (or a prefix of it).
-        node_index = 0
-        n = len(buffer)
-        for k, b in enumerate(buffer):
-            if node_index in self.dead:
-                # This buffer (or a prefix of it) has already been tested,
-                # or has already had its descendants fully explored.
-                # Testing it again would not be helpful.
-                return False
-            try:
-                # The block size at that point provides a lower bound on how
-                # many more bytes are required. If the buffer does not have
-                # enough bytes to fulfill that block size then we can rule out
-                # this buffer.
-                if k + self.block_sizes[node_index] > n:
-                    return False
-            except KeyError:
-                pass
-
-            # If there's a forced value or a mask at this position, then
-            # pretend that the buffer already contains a matching value,
-            # because the test function is going to do the same.
-            try:
-                b = self.forced[node_index]
-            except KeyError:
-                pass
-            try:
-                b = b & self.masks[node_index]
-            except KeyError:
-                pass
-
-            try:
-                node_index = self.tree[node_index][b]
-            except KeyError:
-                # The buffer wasn't in the tree, which means we haven't tried
-                # it. That makes it a possible candidate.
-                return True
-        else:
-            # We ran out of buffer before reaching a leaf or a missing node.
-            # That means the test function is going to draw beyond the end
-            # of this buffer, which makes it a bad candidate.
-            return False
-
     def cached_test_function(self, buffer):
         """Checks the tree to see if we've tested this buffer, and returns the
         previous result if we have.
@@ -1062,43 +791,40 @@ class ConjectureRunner(object):
         Otherwise we call through to ``test_function``, and return a
         fresh result.
         """
-        node_index = 0
-        for c in buffer:
-            # If there's a forced value or a mask at this position, then
-            # pretend that the buffer already contains a matching value,
-            # because the test function is going to do the same.
-            try:
-                c = self.forced[node_index]
-            except KeyError:
-                pass
-            try:
-                c = c & self.masks[node_index]
-            except KeyError:
-                pass
-
-            try:
-                node_index = self.tree[node_index][c]
-            except KeyError:
-                # The byte at this position isn't in the tree, which means
-                # we haven't tested this buffer. Break out of the tree
-                # traversal, and run the test function normally.
-                break
-            node = self.tree[node_index]
-            if isinstance(node, ConjectureData):
-                # This buffer (or a prefix of it) has already been tested.
-                # Return the stored result instead of trying it again.
-                return node
-        else:
-            # Falling off the end of this loop means that we're about to test
-            # a prefix of a previously-tested byte stream. The test is going
-            # to draw beyond the end of the buffer, and fail due to overrun.
-            # Currently there is no special handling for this case.
+        buffer = hbytes(buffer)
+        try:
+            return self.__data_cache[buffer]
+        except KeyError:
             pass
 
+        rewritten, status = self.tree.rewrite(buffer)
+
+        try:
+            result = self.__data_cache[rewritten]
+        except KeyError:
+            pass
+        else:
+            assert result.status != Status.OVERRUN or result is Overrun
+            self.__data_cache[buffer] = result
+            return result
+
         # We didn't find a match in the tree, so we need to run the test
-        # function normally.
-        result = ConjectureData.for_buffer(buffer)
-        self.test_function(result)
+        # function normally. Note that test_function will automatically
+        # add this to the tree so we don't need to update the cache.
+
+        result = None
+
+        if status != Status.OVERRUN:
+            result = ConjectureData.for_buffer(buffer)
+            self.test_function(result)
+            assert status is None or result.status == status
+            status = result.status
+        if status == Status.OVERRUN:
+            result = Overrun
+
+        assert result is not None
+
+        self.__data_cache[buffer] = result
         return result
 
     def event_to_string(self, event):
@@ -1113,16 +839,6 @@ class ConjectureRunner(object):
         return result
 
 
-def _is_simple_mask(mask):
-    """A simple mask is ``(2 ** n - 1)`` for some ``n``, so it has the effect
-    of keeping the lowest ``n`` bits and discarding the rest.
-
-    A mask in this form can produce any integer between 0 and the mask itself
-    (inclusive), and the total number of these values is ``(mask + 1)``.
-    """
-    return (mask & (mask + 1)) == 0
-
-
 def _draw_predecessor(rnd, xs):
     r = bytearray()
     any_strict = False
@@ -1212,12 +928,6 @@ class TargetSelector(object):
             self.best_status = data.status
             self.reset()
 
-        # Note that technically data could be a duplicate. This rarely happens
-        # (only if we've exceeded the CACHE_RESET_FREQUENCY number of test
-        # function calls), but it's certainly possible. This could result in
-        # us having the same example multiple times, possibly spread over both
-        # lists. We could check for this, but it's not a major problem so we
-        # don't bother.
         self.fresh_examples.append(data)
         if len(self) > self.pool_size:
             pop_random(self.random, self.used_examples or self.fresh_examples)
